/****************************************************************************
 *
 * Input.c
 *
 * Source file for handling Input Files.
 ****************************************************************************/

/*
Revisement Log:
     v1.2 2004/09/24, Ray Kuang, Fixed the bug of mapping background frequence 
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "includes.h"

// This file handles I/O for the SST package.  
// It handles printing out input records and predictions.  
// It also handles reading in data from files or standard input.

/**************** Printing Input Records and Predictions *********************/


// printInputRecordInputs() prints out the input sequence
// corresponding to an input record.
void printInputRecordInputs(struct inputRecord *inputRec, 
		      struct symbolTable *inputSymbols, 
		      struct symbolTable *outputSymbols,
		      char *additionalComment) {

  int i;


  for (i=0; i<inputRec->length; i++) {
    printf("%s ", lookUpSymbol(inputSymbols,inputRec->inputs[i]));
  }
  printf("\t%s%s",additionalComment,inputRec->comment);
}

// pintInputRecordOutput() prints out the output symbol of an input record.
void printInputRecordSource(struct inputRecord *inputRec, 
		      struct symbolTable *inputSymbols, 
		      struct symbolTable *outputSymbols,
		      char *additionalComment) {

  printf("%s: ",lookUpSymbol(outputSymbols,inputRec->source));
}


/************** Reading Sequences *****************************************/

/*
int getSequenceLength(FILE *fp) {

  int length=0;
  char tempBuffer[61];

  while (fscanf(fp,"%s",tempBuffer)!=EOF) {
    if (tempBuffer[0]!='>')
      length++;
  }
  return length;
}



int getNumSources(FILE *fp) {

  int sources=0;
  char tempBuffer[61];

  while (fscanf(fp,"%s",tempBuffer)!=EOF) {
    if (tempBuffer[0]=='>')
      sources++;
  }
  return sources;
}

*/

//By Ke Wang
//compute the total number of char in the protein sequences
int getSequenceLength(FILE *fp, int srcNum) {
  int length = 0;
  char tempBuffer[2048];

  while ( fgets(tempBuffer, 2047, fp)!=NULL ){
      length++;
  }

  length = length - 11*srcNum; //11 extra lines for one sequence

  return length;
}


//get the number of the sequence
int getNumSources(FILE *fp){
  int sources=0;
  char tempBuffer[2048];

  while ( fgets(tempBuffer, 2048, fp)!=NULL ){
    if (tempBuffer[0]=='>')
      sources++;

  }
  return sources;
}


/************** Reading Multiple Sequences **********************************/

// TODO RETURN NUM RECORDS THROUGH POINTER AND CHEANGE IN GLOABL SO IT DEOS NOT CRASH.

struct inputRecord **readInputMultSequence(FILE *fp,
					   struct symbolTable *inputSymbols,
					   struct symbolTable *outputSymbols, 
					   int sequenceLength,
					   int sequenceNumber) {

  int i,j,k;
  int numRecords=sequenceLength;
  int *sequence=(int *)calloc(sequenceLength,sizeof(int));
  int *sources=(int *)calloc(sequenceLength,sizeof(int));
  int currentSource=-1;
  int position;
  int *newSequence;
  char tempSymbols[61];
  struct inputRecord **records
    =(struct inputRecord **)calloc(numRecords,sizeof(struct inputRecord *));

  //Ke Wang
  char tempBuffer[2048];
  //char profile[sequenceLength][5];
  //int profileLen[sequenceLength];
  double **profile;
  int *profileLen;
  int profileCount;
  char *index;
  char *letter;
  int candidate[20];//here used to find the corresponding place in the Amino.txt 
  int re_candidate[20];
  char *profileNum[20]; //20 possible profile value
  double background[20]; //the background pdf for the 20 amino acid
  double C=0.8; //for compute the mixture pdf

  int sourceNum = 0;
  
  int allZeroFlag;
  int temp_profile_index;
 /* background pdf
   A       101098          0.0799912015849807
 C       21719   0.0171846021407367
 D       73164   0.0578891399707563
 E       80656   0.0638169929675978
 F       50093   0.0396348024787477
 G       96137   0.0760659374742852
 H       28243   0.0223465499452473
 I       69627   0.0550905793661343
 K       76411   0.060458245507428
 L       109564          0.0866897071203864
 M       27221   0.0215379186368154
 N       55981   0.044293531582512
 P       58864   0.0465746314476874
 Q       48100   0.0380578923048682
 R       61232   0.0484482507611578
 S       79627   0.0630028230885602
 T       73354   0.0580394726014824
 V       88501   0.0700241481678408
 W       18325   0.0144991866213453
 Y       45947   0.03635438623143
  */

  background[0]=0.0799912015849807; //A
  background[1]=0.0484482507611578;//R
  background[2]=0.044293531582512;//N
  background[3]=0.0578891399707563;//D
  background[4]=0.0171846021407367;//C
  background[5]=0.0380578923048682;//Q
  background[6]=0.0638169929675978;//E
  background[7]=0.0760659374742852;//G
  background[8]=0.0223465499452473;//H
  background[9]=0.0550905793661343;//I
  background[10]=0.0866897071203864;//L
  background[11]=0.060458245507428;//K
  background[12]=0.0215379186368154;//M
  background[13]=0.0396348024787477;//F
  background[14]=0.0465746314476874;//P
  background[15]=0.0630028230885602;//S
  background[16]=0.0580394726014824;//T
  background[17]=0.0144991866213453;//W
  background[18]=0.03635438623143;//Y
  background[19]=0.0700241481678408;//V


  candidate[0]=0;//A is 0 place in amino.txt
  candidate[1]=14;//R IS 14 PLACE IN AMINO.TXT
  candidate[2]=11;
  candidate[3]=2;
  candidate[4]=1;
  candidate[5]=13;
  candidate[6]=3;
  candidate[7]=5;
  candidate[8]=6;
  candidate[9]=7;
  candidate[10]=9;
  candidate[11]=8;
  candidate[12]=10;
  candidate[13]=4;
  candidate[14]=12;
  candidate[15]=15;
  candidate[16]=16;
  candidate[17]=18;
  candidate[18]=19;
  candidate[19]=17;
  
  re_candidate[0]=0;//A is 0 place in amino.txt
  re_candidate[14]=1;//R IS 14 PLACE IN AMINO.TXT
  re_candidate[11]=2;
  re_candidate[2]=3;
  re_candidate[1]=4;
  re_candidate[13]=5;
  re_candidate[3]=6;
  re_candidate[5]=7;
  re_candidate[6]=8;
  re_candidate[7]=9;
  re_candidate[9]=10;
  re_candidate[8]=11;
  re_candidate[10]=12;
  re_candidate[4]=13;
  re_candidate[12]=14;
  re_candidate[15]=15;
  re_candidate[16]=16;
  re_candidate[18]=17;
  re_candidate[19]=18;
  re_candidate[17]=19;

  profile = (double **)calloc(sequenceLength, sizeof(double *));
  for(i=0;i<sequenceLength;i++){
    profile[i] = (double *)calloc(20, sizeof(double));//for substitution, should keep all 20 acid
  }
  profileLen = (int *)calloc(sequenceLength, sizeof(int));
  

  //printf("sequence length :%d \n", sequenceLength);

  //Ke Wang
  for (i =0; i<sequenceLength; i++) {

    //fscanf(fp,"%s",tempSymbols);
    fgets(tempBuffer, 1023, fp);

    if (tempBuffer[0]!='>' && strlen(tempBuffer)>60) {
      if (currentSource==-1) {
	fprintf(stderr,"Invalid Input File\n");
	exit(1);
      }

      //printf("%d, %s\n", i,tempBuffer);

      index = strtok(tempBuffer, " ");
      letter = strtok(NULL, " ");
      for(k=0;k<20;k++) {
	strtok(NULL, " "); //skip until the pdf part
      }
      profileCount = 0;
      allZeroFlag=1;
      for(k=0;k<20;k++) {
	profileNum[k] = strtok(NULL, " ");
	profile[i][candidate[k]]=-log(C*(atof(profileNum[k])/100)+(1-C)*background[k]);
	if(atof(profileNum[k])>0)
	   allZeroFlag=0; //not all zero possiblity
	profileCount++;
	
	/*
	if(currentSource==973)
	  printf("first=%f, second=%f, total=%f\n",C*(atof(profileNum[k])/100), (1-C)*background[k], -log(C*(atof(profileNum[k])/100)+(1-C)*background[k]));
	
	if(atoi(profileNum[k])>=THRESHOLD){
	  if(strcmp(candidate[k],letter)!=0){
	    profile[i][profileCount++]=mapSymbol(inputSymbols, candidate[k]);
	  }
	}
	*/
      }
       
      profileLen[i]=profileCount;
      sequence[i]=mapSymbol(inputSymbols,letter);
      sources[i]=currentSource;

      if(allZeroFlag==1){

        if(!(sequence[i]==1 || sequence[i]==20 || sequence[i]==22))
	  //{
	  //for(k=0;k<20;k++)
	  //{
	  //  profile[i][k]=-1000; //put in a large negative
	  //}
	  //}
	  //else
	{
	  if(sequence[i]==21)
	    temp_profile_index=19;
	  else if(sequence[i]>1)
	    temp_profile_index=sequence[i]-1;
	  else
	    temp_profile_index=sequence[i];
	  profile[i][temp_profile_index]=-log(C+(1-C)*background[re_candidate[temp_profile_index]]);
	}

	//printf("all zero seq: %d\n", currentSource);
	//printf("pos %d, value %f\n", sequence[i], profile[i][sequence[i]]);
      }

      //printf("%d, %s\n",i,letter);

    } else if(tempBuffer[0]!='>'){ //skip the end part of a sequence
      
      for(k=0;k<5;k++){
	fgets(tempBuffer, 1023, fp);
      }
      
      i--;

    } else { //the line starts with >

      sscanf(tempBuffer, "%s", tempSymbols); //get the first string out
      //printf("ID is: %s\n", tempSymbols);
      //printf("i=%d\n", i);
      //currentSource=mapSymbol(outputSymbols,tempSymbols+1);
      currentSource = sourceNum;
      sourceNum++;
      //printf("ID is: %s, source is %d\n", tempSymbols,currentSource);

      numRecords-=MAX_DEPTH-1;
      if (i>MAX_DEPTH) {
	for (j=0; j<MAX_DEPTH-1; j++) {
	  sources[i-j-1]=-1;
	}
      }
      i--;

      for(k=0;k<4;k++){
	fgets(tempBuffer, 2047, fp); //skip next 4 useless lines
      }
      //printf("done..\n");

    } //end of else

    //printf("%d,",i);
  } //end of for i

  if (i>MAX_DEPTH) {
    for (j=0; j<MAX_DEPTH-1; j++) {
      sources[i-j-1]=-1;
    }
  }
  
  //printf("finish reading the files...\n");

  newSequence=(int *)calloc(sequenceLength+globalSequenceLength,sizeof(int));
  for (i=0; i<globalSequenceLength; i++) {
    newSequence[i]=globalSequence[i];
  }
  for (i=globalSequenceLength; i<globalSequenceLength+sequenceLength; i++) {
    newSequence[i]=sequence[i-globalSequenceLength];
  }
  if (globalSequenceLength>0) {
    free(globalSequence);
  }
  globalSequence=newSequence;

  //Ke
  globalProfile = profile;
  globalProfileLen = profileLen;

  j=0;
  position=0;
  for (i=0; i<numRecords; i++) {
    while (sources[j]==-1) {
      j++;
      position=0;
    }
    position++;

    records[i]=(struct inputRecord *)malloc(sizeof(struct inputRecord));
    records[i]->source=sources[j];
    records[i]->inputs=&sequence[j];
    records[i]->position=position;
    records[i]->length=MAX_DEPTH;
    records[i]->comment=NULL;

    //Ke
    records[i]->profilelen = &profileLen[j];
    records[i]->profile = &profile[j];

    j++;

    //keke
    /*
    printf("%d record:", i);
    printf("input %s%s%s ",lookUpSymbol(inputSymbols, records[i]->inputs[0]),lookUpSymbol(inputSymbols, records[i]->inputs[1]),lookUpSymbol(inputSymbols, records[i]->inputs[2]));
    printf("source %s ", lookUpSymbol(outputSymbols, records[i]->source));
    printf("position %d\n", records[i]->position);
    */

  }
  
  
  globalSequenceLength+=sequenceLength;


  return records;
}

struct inputRecord **readMultipleSequences(char **fileNames,
					   int numFiles,
					   struct symbolTable *inputSymbols, 
					   struct symbolTable *outputSymbols) {
  
  int sequenceLength=0;
  struct inputRecord **records,**newRecords,**oldRecords;
  int i,j;
  int numSources,numRecords;
  FILE *fp;

  numTotalRecords=0;

  globalSequenceLength=0;


  for (j=0; j<numFiles; j++) {

    if (VERBOSE)
      printf("Current File Name = %s\n",fileNames[j]);

    
    if ((fp = fopen(fileNames[j],"r"))==NULL) {
      fprintf(stderr, "Invalid Input Sequence File:%s\n",fileNames[j]);
      exit(1);
    }
    
    numSources=getNumSources(fp);
    fclose(fp);

    //printf("numSrc=%d \n",numSources);
    
    if ((fp = fopen(fileNames[j],"r"))==NULL) {
      fprintf(stderr, "Invalid Input Sequence File:%s\n",fileNames[j]);
      exit(1);
    }
    
    sequenceLength=getSequenceLength(fp, numSources);
    fclose(fp);
    
    //printf("the seqlength=%d, ", sequenceLength);

    
    
    if ((fp = fopen(fileNames[j],"r"))==NULL) {
      fprintf(stderr, "Invalid Input Sequence File:%s\n",fileNames[j]);
      exit(1);
    }

    numRecords=sequenceLength-(numSources*(MAX_DEPTH-1));
    records=readInputMultSequence(fp,inputSymbols,outputSymbols,sequenceLength,j);
    
    oldRecords=newRecords;
    newRecords=calloc(numTotalRecords+numRecords,sizeof(struct inputRecord *));
    for (i=0; i<numTotalRecords; i++) {
      newRecords[i]=oldRecords[i];
    }
    if (numTotalRecords>0)
      free(oldRecords);
    
    
    for (i=numTotalRecords; i<numTotalRecords+numRecords; i++) {
      
      newRecords[i]=records[i-numTotalRecords];
      if (VERBOSE && i % 1000==0) {
	printf("Reading Example %d\n",i);
      }
    }
    free(records);
    
    numTotalRecords+=numRecords;
    
  }

  return newRecords;

  //return NULL;
} 



